@@ -43,6 +43,7 @@
 
 // @snippet include-pyside
 #include <pyside.h>
+#include <limits>
 // @snippet include-pyside
 
 // @snippet pystring-check
@@ -63,14 +64,19 @@ bool py2kStrCheck(PyObject *obj)
 // defaultValue can also be passed as positional argument,
 // not only as keyword.
 QVariant out;
-if (kwds || numArgs > 1)
+if (kwds || numArgs > 1) {
+    Py_BEGIN_ALLOW_THREADS
     out = %CPPSELF.value(%1, %2);
-else
+    Py_END_ALLOW_THREADS
+} else {
+    Py_BEGIN_ALLOW_THREADS
     out = %CPPSELF.value(%1);
+    Py_END_ALLOW_THREADS
+}
 
 PyTypeObject *typeObj = reinterpret_cast<PyTypeObject*>(%PYARG_3);
 
-if (typeObj) {
+if (typeObj && !Shiboken::ObjectType::checkType(typeObj)) {
     if (typeObj == &PyList_Type) {
         QByteArray out_ba = out.toByteArray();
         if (!out_ba.isEmpty()) {
@@ -116,8 +122,14 @@ if (typeObj) {
             Py_INCREF(Py_False);
             %PYARG_0 = Py_False;
         }
+    } else {
+        // TODO: PyDict_Type and PyTuple_Type
+        PyErr_SetString(PyExc_TypeError,
+                        "Invalid type parameter.\n"
+                        "\tUse 'list', 'bytes', 'str', 'int', 'float', 'bool', "
+                        "or a Qt-derived type");
+        return nullptr;
     }
-    // TODO: PyDict_Type and PyTuple_Type
 }
 else {
     if (!out.isValid()) {
@@ -368,9 +380,15 @@ static bool getReceiver(QObject *source, const char *signal, PyObject *callback,
             usingGlobalReceiver = true;
     }
 
+    const auto receiverThread = *receiver ? (*receiver)->thread() : nullptr;
+
     if (usingGlobalReceiver) {
         PySide::SignalManager &signalManager = PySide::SignalManager::instance();
         *receiver = signalManager.globalReceiver(source, callback);
+        // PYSIDE-1354: Move the global receiver to the original receivers's thread
+        // so that autoconnections work correctly.
+        if (receiverThread && receiverThread != (*receiver)->thread())
+            (*receiver)->moveToThread(receiverThread);
         *callbackSig = PySide::Signal::getCallbackSignature(signal, *receiver, callback, usingGlobalReceiver).toLatin1();
     }
 
@@ -393,9 +411,7 @@ static bool qobjectConnect(QObject *source, const char *signal, QObject *receive
     slot++;
     PySide::SignalManager::registerMetaMethod(receiver, slot, isSignal ? QMetaMethod::Signal : QMetaMethod::Slot);
     bool connection;
-    Py_BEGIN_ALLOW_THREADS
     connection = QObject::connect(source, signal - 1, receiver, slot - 1, type);
-    Py_END_ALLOW_THREADS
     return connection;
 }
 
@@ -451,9 +467,7 @@ static bool qobjectConnectCallback(QObject *source, const char *signal, PyObject
         }
     }
     bool connection;
-    Py_BEGIN_ALLOW_THREADS
     connection = QMetaObject::connect(source, signalIndex, receiver, slotIndex, type);
-    Py_END_ALLOW_THREADS
     if (connection) {
         if (usingGlobalReceiver)
             signalManager.notifyGlobalReceiver(receiver);
@@ -499,9 +513,7 @@ static bool qobjectDisconnectCallback(QObject *source, const char *signal, PyObj
     slotMethod = metaObject->method(slotIndex);
 
     bool disconnected;
-    Py_BEGIN_ALLOW_THREADS
     disconnected = QMetaObject::disconnectOne(source, signalIndex, receiver, slotIndex);
-    Py_END_ALLOW_THREADS
 
     if (disconnected) {
         if (usingGlobalReceiver)
@@ -577,14 +589,28 @@ bool %0 = qobjectConnect(%1, %2, %3, %4, %5);
 // qFatal doesn't have a stream version, so we do a
 // qWarning call followed by a qFatal() call using a
 // literal.
+Py_BEGIN_ALLOW_THREADS
 qWarning() << %1;
 qFatal("[A qFatal() call was made from Python code]");
+Py_END_ALLOW_THREADS
 // @snippet qfatal
 
 // @snippet moduleshutdown
 PySide::runCleanupFunctions();
 // @snippet moduleshutdown
 
+// @snippet qt-qenum
+%PYARG_0 = PySide::QEnum::QEnumMacro(%1, false);
+// @snippet qt-qenum
+
+// @snippet qt-qflag
+%PYARG_0 = PySide::QEnum::QEnumMacro(%1, true);
+// @snippet qt-qflag
+
+// @snippet qt-init-feature
+PySide::Feature::init();
+// @snippet qt-init-feature
+
 // @snippet qt-pysideinit
 Shiboken::Conversions::registerConverterName(SbkPySide2_QtCoreTypeConverters[SBK_QSTRING_IDX], "unicode");
 Shiboken::Conversions::registerConverterName(SbkPySide2_QtCoreTypeConverters[SBK_QSTRING_IDX], "str");
@@ -606,7 +632,7 @@ static void msgHandlerCallback(QtMsgType type, const QMessageLogContext &ctx, co
     Shiboken::AutoDecRef arglist(PyTuple_New(3));
     PyTuple_SET_ITEM(arglist, 0, %CONVERTTOPYTHON[QtMsgType](type));
     PyTuple_SET_ITEM(arglist, 1, %CONVERTTOPYTHON[QMessageLogContext &](ctx));
-    QByteArray array = msg.toLocal8Bit();
+    QByteArray array = msg.toUtf8();  // Python handler requires UTF-8
     char *data = array.data();
     PyTuple_SET_ITEM(arglist, 2, %CONVERTTOPYTHON[char *](data));
     Shiboken::AutoDecRef ret(PyObject_CallObject(qtmsghandler, arglist));
@@ -637,7 +663,7 @@ if (%PYARG_0 == Py_None)
 
 // @snippet qline-hash
 namespace PySide {
-    template<> inline uint hash(const QLine &v) {
+    template<> inline Py_ssize_t hash(const QLine &v) {
         return qHash(qMakePair(qMakePair(v.x1(), v.y1()), qMakePair(v.x2(), v.y2())));
     }
 };
@@ -669,9 +695,7 @@ if (!PyDateTimeAPI)
 
 // @snippet qdate-getdate
 int year, month, day;
-%BEGIN_ALLOW_THREADS
 %CPPSELF.%FUNCTION_NAME(&year, &month, &day);
-%END_ALLOW_THREADS
 %PYARG_0 = PyTuple_New(3);
 PyTuple_SET_ITEM(%PYARG_0, 0, %CONVERTTOPYTHON[int](year));
 PyTuple_SET_ITEM(%PYARG_0, 1, %CONVERTTOPYTHON[int](month));
@@ -680,9 +704,7 @@ PyTuple_SET_ITEM(%PYARG_0, 2, %CONVERTTOPYTHON[int](day));
 
 // @snippet qdate-weeknumber
 int yearNumber;
-%BEGIN_ALLOW_THREADS
 int week = %CPPSELF.%FUNCTION_NAME(&yearNumber);
-%END_ALLOW_THREADS
 %PYARG_0 = PyTuple_New(2);
 PyTuple_SET_ITEM(%PYARG_0, 0, %CONVERTTOPYTHON[int](week));
 PyTuple_SET_ITEM(%PYARG_0, 1, %CONVERTTOPYTHON[int](yearNumber));
@@ -709,7 +731,7 @@ if (!PyDateTimeAPI) PySideDateTime_IMPORT;
 
 // @snippet qpoint
 namespace PySide {
-    template<> inline uint hash(const QPoint &v) {
+    template<> inline Py_ssize_t hash(const QPoint &v) {
         return qHash(qMakePair(v.x(), v.y()));
     }
 };
@@ -717,7 +739,7 @@ namespace PySide {
 
 // @snippet qrect
 namespace PySide {
-    template<> inline uint hash(const QRect &v) {
+    template<> inline Py_ssize_t hash(const QRect &v) {
         return qHash(qMakePair(qMakePair(v.x(), v.y()), qMakePair(v.width(), v.height())));
     }
 };
@@ -725,7 +747,7 @@ namespace PySide {
 
 // @snippet qsize
 namespace PySide {
-    template<> inline uint hash(const QSize &v) {
+    template<> inline Py_ssize_t hash(const QSize &v) {
         return qHash(qMakePair(v.width(), v.height()));
     }
 };
@@ -777,12 +799,35 @@ qRegisterMetaType<QVector<int> >("QVector<int>");
 // @snippet qobject-metaobject
 
 // @snippet qobject-findchild-1
+static bool _findChildTypeMatch(const QObject *child, PyTypeObject *desiredType)
+{
+    auto *pyChildType = PySide::getTypeForQObject(child);
+    return pyChildType != nullptr && PyType_IsSubtype(pyChildType, desiredType);
+}
+
+static inline bool _findChildrenComparator(const QObject *child,
+                                           const QRegularExpression &name)
+{
+    return name.match(child->objectName()).hasMatch();
+}
+
+static inline bool _findChildrenComparator(const QObject *child,
+                                           const QString &name)
+{
+    return name.isNull() || name == child->objectName();
+}
+
+static inline bool _findChildrenComparator(const QObject *child,
+                                           const QRegExp &name)
+{
+    return name.indexIn(child->objectName()) != -1;
+}
+
 static QObject *_findChildHelper(const QObject *parent, const QString &name, PyTypeObject *desiredType)
 {
     for (auto *child : parent->children()) {
-        Shiboken::AutoDecRef pyChild(%CONVERTTOPYTHON[QObject *](child));
-        if (PyType_IsSubtype(Py_TYPE(pyChild), desiredType)
-            && (name.isNull() || name == child->objectName())) {
+        if (_findChildrenComparator(child, name)
+            && _findChildTypeMatch(child, desiredType)) {
             return child;
         }
     }
@@ -795,23 +840,15 @@ static QObject *_findChildHelper(const QObject *parent, const QString &name, PyT
     return nullptr;
 }
 
-static inline bool _findChildrenComparator(const QObject *&child, const QRegExp &name)
-{
-    return name.indexIn(child->objectName()) != -1;
-}
-
-static inline bool _findChildrenComparator(const QObject *&child, const QString &name)
-{
-    return name.isNull() || name == child->objectName();
-}
-
-template<typename T>
+template<typename T> // QString/QRegularExpression/QRegExp
 static void _findChildrenHelper(const QObject *parent, const T& name, PyTypeObject *desiredType, PyObject *result)
 {
     for (const auto *child : parent->children()) {
-        Shiboken::AutoDecRef pyChild(%CONVERTTOPYTHON[QObject *](child));
-        if (PyType_IsSubtype(Py_TYPE(pyChild), desiredType) && _findChildrenComparator(child, name))
+        if (_findChildrenComparator(child, name)
+            && _findChildTypeMatch(child, desiredType)) {
+            Shiboken::AutoDecRef pyChild(%CONVERTTOPYTHON[QObject *](child));
             PyList_Append(result, pyChild);
+        }
         _findChildrenHelper(child, name, desiredType, result);
     }
 }
@@ -822,29 +859,46 @@ QObject *child = _findChildHelper(%CPPSELF, %2, reinterpret_cast<PyTypeObject *>
 %PYARG_0 = %CONVERTTOPYTHON[QObject *](child);
 // @snippet qobject-findchild-2
 
-// @snippet qobject-findchildren-1
+// @snippet qobject-findchildren
 %PYARG_0 = PyList_New(0);
 _findChildrenHelper(%CPPSELF, %2, reinterpret_cast<PyTypeObject *>(%PYARG_1), %PYARG_0);
-// @snippet qobject-findchildren-1
+// @snippet qobject-findchildren
 
-// @snippet qobject-findchildren-2
-%PYARG_0 = PyList_New(0);
-_findChildrenHelper(%CPPSELF, %2, reinterpret_cast<PyTypeObject *>(%PYARG_1), %PYARG_0);
-// @snippet qobject-findchildren-2
+//////////////////////////////////////////////////////////////////////////////
+// PYSIDE-131: Use the class name as context where the calling function is
+//             living. Derived Python classes have the wrong context.
+//
+// The original patch uses Python introspection to look up the current
+// function (from the frame stack) in the class __dict__ along the mro.
+//
+// The problem is that looking into the frame stack works for Python
+// functions, only. For including builtin function callers, the following
+// approach turned out to be much simpler:
+//
+// Walk the __mro__
+// - translate the string
+// - if the translated string is changed:
+//   - return the translation.
 
 // @snippet qobject-tr
-QString result;
-if (QCoreApplication::instance()) {
-    PyObject *klass = PyObject_GetAttr(%PYSELF, Shiboken::PyMagicName::class_());
-    PyObject *cname = PyObject_GetAttr(klass, Shiboken::PyMagicName::name());
-    result = QString(QCoreApplication::instance()->translate(Shiboken::String::toCString(cname),
-                                                        /*   %1, %2, QCoreApplication::CodecForTr, %3)); */
-                                                             %1, %2, %3));
-
-    Py_DECREF(klass);
-    Py_DECREF(cname);
-} else {
-    result = QString(QString::fromLatin1(%1));
+PyTypeObject *type = Py_TYPE(%PYSELF);
+PyObject *mro = type->tp_mro;
+auto len = PyTuple_GET_SIZE(mro);
+QString result = QString::fromUtf8(%1);
+QString oldResult = result;
+static auto *sbkObjectType = reinterpret_cast<PyTypeObject *>(SbkObject_TypeF());
+for (Py_ssize_t idx = 0; idx < len - 1; ++idx) {
+    // Skip the last class which is `object`.
+    auto *type = reinterpret_cast<PyTypeObject *>(PyTuple_GET_ITEM(mro, idx));
+    if (type == sbkObjectType)
+        continue;
+    const char *context = type->tp_name;
+    const char *dotpos = strrchr(context, '.');
+    if (dotpos != nullptr)
+        context = dotpos + 1;
+    result = QCoreApplication::translate(context, %1, %2, %3);
+    if (result != oldResult)
+        break;
 }
 %PYARG_0 = %CONVERTTOPYTHON[QString](result);
 // @snippet qobject-tr
@@ -864,7 +918,7 @@ if (ret > 0 && ((strcmp(%1, SIGNAL(destroyed())) == 0) || (strcmp(%1, SIGNAL(des
 // @snippet qregexp-replace
 
 // @snippet qbytearray-mgetitem
-if (PyIndex_Check(_key)) {
+if (PepIndex_Check(_key)) {
     Py_ssize_t _i;
     _i = PyNumber_AsSsize_t(_key, PyExc_IndexError);
     if (_i < 0 || _i >= %CPPSELF.size()) {
@@ -915,7 +969,7 @@ if (PyIndex_Check(_key)) {
 // @snippet qbytearray-mgetitem
 
 // @snippet qbytearray-msetitem
-if (PyIndex_Check(_key)) {
+if (PepIndex_Check(_key)) {
     Py_ssize_t _i = PyNumber_AsSsize_t(_key, PyExc_IndexError);
     if (_i == -1 && PyErr_Occurred())
         return -1;
@@ -1011,9 +1065,10 @@ if (PyIndex_Check(_key)) {
             if (PyLong_Check(item) || PyInt_Check(item)) {
 #endif
                 int overflow;
-                long ival = PyLong_AsLongAndOverflow(item, &overflow);
-                // Not suppose to bigger than 255 because only bytes, bytearray, QByteArray were accept
-                temp = QByteArray(reinterpret_cast<const char *>(&ival));
+                const long ival = PyLong_AsLongAndOverflow(item, &overflow);
+                // Not supposed to be bigger than 255 because only bytes,
+                // bytearray, QByteArray were accepted
+                temp.append(char(ival));
             } else {
                 temp = %CONVERTTOCPP[QByteArray](item);
             }
@@ -1046,6 +1101,7 @@ static int SbkQByteArray_getbufferproc(PyObject *obj, Py_buffer *view, int flags
 
     QByteArray * cppSelf = %CONVERTTOCPP[QByteArray *](obj);
     //XXX      /|\ omitting this space crashes shiboken!
+ #ifdef Py_LIMITED_API
     view->obj = obj;
     view->buf = reinterpret_cast<void *>(cppSelf->data());
     view->len = cppSelf->size();
@@ -1053,13 +1109,20 @@ static int SbkQByteArray_getbufferproc(PyObject *obj, Py_buffer *view, int flags
     view->itemsize = 1;
     view->format = const_cast<char *>("c");
     view->ndim = 1;
-    view->shape = NULL;
+    view->shape = (flags & PyBUF_ND) == PyBUF_ND ? &(view->len) : nullptr;
     view->strides = &view->itemsize;
     view->suboffsets = NULL;
     view->internal = NULL;
 
     Py_XINCREF(obj);
     return 0;
+#else // Py_LIMITED_API
+    const int result = PyBuffer_FillInfo(view, obj, reinterpret_cast<void *>(cppSelf->data()),
+                                         cppSelf->size(), 0, flags);
+    if (result == 0)
+        Py_XINCREF(obj);
+    return result;
+#endif
 }
 
 #if PY_VERSION_HEX < 0x03000000
@@ -1282,6 +1345,14 @@ QByteArray ba(1 + int(%2), char(0));
 %CPPSELF.%FUNCTION_NAME(Shiboken::String::toCString(%PYARG_1), Shiboken::String::len(%PYARG_1));
 // @snippet qcryptographichash-adddata
 
+// @snippet qsocketdescriptor
+#ifdef WIN32
+using DescriptorType = Qt::HANDLE;
+#else
+using DescriptorType = int;
+#endif
+// @snippet qsocketdescriptor
+
 // @snippet qsocketnotifier
 PyObject *socket = %PYARG_1;
 if (socket != nullptr) {
@@ -1422,8 +1493,8 @@ if (qApp) {
         // this will keep app live after python exit (extra ref)
 }
 // PYSIDE-571: make sure that we return the singleton "None"
-if (pyApp == Py_None)
-    Py_DECREF(MakeSingletonQAppWrapper(nullptr)); // here qApp and instance() diverge
+if (Py_TYPE(pyApp) == Py_TYPE(Py_None))
+    Py_DECREF(MakeQAppWrapper(nullptr));
 %PYARG_0 = pyApp;
 Py_XINCREF(%PYARG_0);
 // @snippet qcoreapplication-instance
@@ -1431,7 +1502,10 @@ Py_XINCREF(%PYARG_0);
 // @snippet qdatastream-readrawdata
 QByteArray data;
 data.resize(%2);
-int result = %CPPSELF.%FUNCTION_NAME(data.data(), data.size());
+int result = 0;
+Py_BEGIN_ALLOW_THREADS
+result = %CPPSELF.%FUNCTION_NAME(data.data(), data.size());
+Py_END_ALLOW_THREADS
 if (result == -1) {
     Py_INCREF(Py_None);
     %PYARG_0 = Py_None;
@@ -1441,7 +1515,10 @@ if (result == -1) {
 // @snippet qdatastream-readrawdata
 
 // @snippet qdatastream-writerawdata
-int r = %CPPSELF.%FUNCTION_NAME(%1, Shiboken::String::len(%PYARG_1));
+int r = 0;
+Py_BEGIN_ALLOW_THREADS
+r = %CPPSELF.%FUNCTION_NAME(%1, Shiboken::String::len(%PYARG_1));
+Py_END_ALLOW_THREADS
 %PYARG_0 = %CONVERTTOPYTHON[int](r);
 // @snippet qdatastream-writerawdata
 
@@ -1572,7 +1649,9 @@ QT_END_NAMESPACE
 // @snippet use-stream-for-format-security
 // Uses the stream version for security reasons
 // see gcc man page at -Wformat-security
+Py_BEGIN_ALLOW_THREADS
 %FUNCTION_NAME() << %1;
+Py_END_ALLOW_THREADS
 // @snippet use-stream-for-format-security
 
 // @snippet qresource-registerResource
@@ -1586,12 +1665,16 @@ QT_END_NAMESPACE
 // @snippet qstring-return
 
 // @snippet stream-write-method
+Py_BEGIN_ALLOW_THREADS
 (*%CPPSELF) << %1;
+Py_END_ALLOW_THREADS
 // @snippet stream-write-method
 
 // @snippet stream-read-method
 %RETURN_TYPE _cpp_result;
+Py_BEGIN_ALLOW_THREADS
 (*%CPPSELF) >> _cpp_result;
+Py_END_ALLOW_THREADS
 %PYARG_0 = %CONVERTTOPYTHON[%RETURN_TYPE](_cpp_result);
 // @snippet stream-read-method
 
@@ -1613,7 +1696,9 @@ if (PyBytes_Check(%PYARG_0)) {
 
 // @snippet qiodevice-readData
 QByteArray ba(1 + int(%2), char(0));
+Py_BEGIN_ALLOW_THREADS
 %CPPSELF.%FUNCTION_NAME(ba.data(), int(%2));
+Py_END_ALLOW_THREADS
 %PYARG_0 = Shiboken::String::fromCString(ba.constData());
 // @snippet qiodevice-readData
 
@@ -1658,20 +1743,50 @@ QByteArray ba(1 + int(%2), char(0));
 %out = %OUTTYPE(PyLong_AsUnsignedLong(%in));
 // @snippet conversion-pylong-unsigned
 
-// @snippet conversion-pyunicode
-#ifndef Py_LIMITED_API
-Py_UNICODE *unicode = PyUnicode_AS_UNICODE(%in);
-# if defined(Py_UNICODE_WIDE)
-// cast as Py_UNICODE can be a different type
-%out = QString::fromUcs4((const uint *)unicode);
-# else
-%out = QString::fromUtf16((const ushort *)unicode, PyUnicode_GET_SIZE(%in));
-# endif
+// @snippet conversion-pylong-quintptr
+#if defined(IS_PY3K) && QT_POINTER_SIZE == 8
+%out = %OUTTYPE(PyLong_AsUnsignedLongLong(%in));
 #else
-wchar_t *temp = PyUnicode_AsWideCharString(%in, NULL);
-%out = QString::fromWCharArray(temp);
-PyMem_Free(temp);
+%out = %OUTTYPE(PyLong_AsUnsignedLong(%in));
 #endif
+// @snippet conversion-pylong-quintptr
+
+// @snippet conversion-pyunicode
+#if defined(Py_LIMITED_API)
+    wchar_t *temp = PyUnicode_AsWideCharString(%in, NULL);
+    %out = QString::fromWCharArray(temp);
+    PyMem_Free(temp);
+#elif defined(IS_PY3K)
+    void *data = PyUnicode_DATA(%in);
+    Py_ssize_t len = PyUnicode_GetLength(%in);
+    switch (PyUnicode_KIND(%in)) {
+        case PyUnicode_1BYTE_KIND:
+            %out = QString::fromLatin1(reinterpret_cast<const char *>(data));
+            break;
+        case PyUnicode_2BYTE_KIND:
+            %out = QString::fromUtf16(reinterpret_cast<const char16_t *>(data), len);
+            break;
+        case PyUnicode_4BYTE_KIND:
+            %out = QString::fromUcs4(reinterpret_cast<const char32_t *>(data), len);
+            break;
+    }
+#else // IS_PY3K
+Py_UNICODE *unicode = PyUnicode_AS_UNICODE(%in);
+#  if defined(Py_UNICODE_WIDE)
+// cast as Py_UNICODE can be a different type
+#    if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+%out = QString::fromUcs4(reinterpret_cast<const char32_t *>(unicode));
+#    else
+%out = QString::fromUcs4(reinterpret_cast<const uint *>(unicode));
+#    endif // Qt 6
+#  else // Py_UNICODE_WIDE
+#    if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+%out = QString::fromUtf16(reinterpret_cast<const char16_t *>(unicode), PepUnicode_GetLength(%in));
+#    else
+%out = QString::fromUtf16(reinterpret_cast<const ushort *>(unicode), PepUnicode_GetLength(%in));
+#    endif // Qt 6
+# endif
+#endif // !IS_PY3K
 // @snippet conversion-pyunicode
 
 // @snippet conversion-pystring
@@ -1697,8 +1812,11 @@ int i = %CONVERTTOCPP[int](%in);
 // @snippet conversion-pyint
 
 // @snippet conversion-qlonglong
+// PYSIDE-1250: For QVariant, if the type fits into an int; use int preferably.
 qlonglong in = %CONVERTTOCPP[qlonglong](%in);
-%out = %OUTTYPE(in);
+constexpr qlonglong intMax = qint64(std::numeric_limits<int>::max());
+constexpr qlonglong intMin = qint64(std::numeric_limits<int>::min());
+%out = in >= intMin && in <= intMax ? %OUTTYPE(int(in)) : %OUTTYPE(in);
 // @snippet conversion-qlonglong
 
 // @snippet conversion-qstring
@@ -1847,6 +1965,14 @@ return PyLong_FromLong(%in);
 return PyLong_FromUnsignedLong(%in);
 // @snippet return-pylong-unsigned
 
+// @snippet return-pylong-quintptr
+#if defined(IS_PY3K) && QT_POINTER_SIZE == 8
+return PyLong_FromUnsignedLongLong(%in);
+#else
+return PyLong_FromUnsignedLong(%in);
+#endif
+// @snippet return-pylong-quintptr
+
 // @snippet return-pyunicode
 QByteArray ba = %in.toUtf8();
 return PyUnicode_FromStringAndSize(ba.constData(), ba.size());
@@ -1921,3 +2047,41 @@ PyTuple_SET_ITEM(%out, 0, %CONVERTTOPYTHON[%INTYPE_0](%in.first));
 PyTuple_SET_ITEM(%out, 1, %CONVERTTOPYTHON[%INTYPE_1](%in.second));
 return %out;
 // @snippet return-qpair
+
+// @snippet qthread_pthread_cleanup
+#ifdef Q_OS_UNIX
+#  include <stdio.h>
+#  include <pthread.h>
+static void qthread_pthread_cleanup(void *arg)
+{
+    // PYSIDE 1282: When terminating a thread using QThread::terminate()
+    // (pthread_cancel()), QThread::run() is aborted and the lock is released,
+    // but ~GilState() is still executed for some reason. Prevent it from
+    // releasing.
+    auto gil = reinterpret_cast<Shiboken::GilState *>(arg);
+    gil->abandon();
+}
+#endif // Q_OS_UNIX
+// @snippet qthread_pthread_cleanup
+
+// @snippet qthread_pthread_cleanup_install
+#ifdef Q_OS_UNIX
+pthread_cleanup_push(qthread_pthread_cleanup, &gil);
+#endif
+// @snippet qthread_pthread_cleanup_install
+
+// @snippet qthread_pthread_cleanup_uninstall
+#ifdef Q_OS_UNIX
+pthread_cleanup_pop(0);
+#endif
+// @snippet qthread_pthread_cleanup_uninstall
+
+// @snippet qlibraryinfo_build
+#if defined(IS_PY3K) && defined(Py_LIMITED_API)
+auto suffix = PyUnicode_FromString(" [limited API]");
+auto oldResult = pyResult;
+pyResult = PyUnicode_Concat(pyResult, suffix);
+Py_DECREF(oldResult);
+Py_DECREF(suffix);
+#endif
+// @snippet qlibraryinfo_build
